"""
Wrapper (with monkeypatched behaviour) for original pycsw WSGI application object
"""

# Imports from . import * are needed to override original PyCSW behaviour
# pylint: disable=unused-import

import sys

from pycsw.wsgi import application

# pylint: disable=no-name-in-module
from . import apiso, config, metadata, repository, server

if __name__ == "__main__":  # run inline using WSGI reference implementation

    from wsgiref.simple_server import make_server

    PORT = 8000
    if len(sys.argv) > 1:
        PORT = int(sys.argv[1])
    httpd = make_server("", PORT, application)
    print(f"Serving on port {PORT}")
    httpd.serve_forever()



"""
Monkeypatch for pycsw.core.admin module
"""

# Imports from . import * are needed to override original PyCSW behaviour
# pylint: disable=C,W,R

from shutil import which

import pycsw.core.admin
from sqlalchemy import Column, Text

# pylint: disable=no-name-in-module
from . import apiso, config, metadata, repository, server

setup_db = pycsw.core.admin.setup_db


def rpi_setup_db(
    database,
    table,
    home,
    create_sfsql_tables=True,
    create_plpythonu_functions=True,
    postgis_geometry_column="wkb_geometry",
    extra_columns=[],
    language="english",
):
    """
    Patched version of pycsw.core.admin.setup_db function
    """

    extra_columns += [
        Column("rpi_organization_uuid", Text, index=True),
        Column("rpi_record_uuid", Text, index=True),
        Column("rpi_is_viewable", Text, index=True),
        Column("rpi_is_searchable", Text, index=True),
    ]

    setup_db(
        database,
        table,
        home,
        create_sfsql_tables,
        create_plpythonu_functions,
        postgis_geometry_column,
        extra_columns,
        language,
    )


pycsw.core.admin.setup_db = rpi_setup_db


def run():
    pycsw_admin_script = which("pycsw-admin.py")
    if pycsw_admin_script:
        exec(
            open(pycsw_admin_script).read().replace("pycsw-admin.py", "pycsw_rpi-admin")
        )
    else:
        raise FileExistsError("Pycsw admin script 'pycsw-admin.py' not found in $PATH")



"""
Monkeypatches for pycsw.core.config module
"""


import pycsw.core.config

staticcontenxt__init__ = pycsw.core.config.StaticContext.__init__


def rpi_staticcontext__init__(self, prefix="csw30"):

    """
    Patched version of pycsw.core.config.StaticContext.__init__ function
    """

    staticcontenxt__init__(self, prefix)
    self.md_core_model["mappings"]["pycsw:RPIRecordUUID"] = "rpi_record_uuid"
    self.md_core_model["mappings"][
        "pycsw:RPIOrganizationUUID"
    ] = "rpi_organization_uuid"
    self.md_core_model["mappings"]["pycsw:RPIIsViewable"] = "rpi_is_viewable"
    self.md_core_model["mappings"]["pycsw:RPIIsSearchable"] = "rpi_is_searchable"
    self.namespaces["rpi"] = "https://rpi.gov.sk/"


pycsw.core.config.StaticContext.__init__ = rpi_staticcontext__init__



"""
Monkeypatches for pycsw.core.apiso.apiso module
"""


import pycsw.plugins.profiles.apiso.apiso

apiso__init__ = pycsw.plugins.profiles.apiso.apiso.APISO.__init__


def rpi_apiso__init__(self, model, namespaces, context):
    """
    Patched version of pycsw.plugins.profiles.apiso.apiso.APISO.__init__ function
    """

    apiso__init__(self, model, namespaces, context)
    self.repository["queryables"]["RPIQueryables"] = {
        "rpi:OrganizationUUID": {
            # "xpath": "//",
            "dbcol": self.context.md_core_model["mappings"][
                "pycsw:RPIOrganizationUUID"
            ],
        },
        "rpi:RecordUUID": {
            # "xpath": "//",
            "dbcol": self.context.md_core_model["mappings"]["pycsw:RPIRecordUUID"],
        },
        "rpi:IsViewable": {
            # "xpath": "//",
            "dbcol": self.context.md_core_model["mappings"]["pycsw:RPIIsViewable"],
        },
        "rpi:IsSearchable": {
            # "xpath": "//",
            "dbcol": self.context.md_core_model["mappings"]["pycsw:RPIIsSearchable"],
        },
    }


pycsw.plugins.profiles.apiso.apiso.APISO.__init__ = rpi_apiso__init__



"""
Monkeypatches for pycsw.server module
"""


import pycsw.server

Csw__init__ = pycsw.server.Csw.__init__


def rpi_csw__init__(self, rtconfig=None, env=None, version="3.0.0"):

    """
    Patched version of pycsw.server.Csw.__init__ function
    """

    Csw__init__(self, rtconfig, env, version)

    self.context.transactions_webhook_url = self.config.get(
        "manager", "transactions_webhook_url", fallback=None
    )


pycsw.server.Csw.__init__ = rpi_csw__init__



"""
Monkeypatches for pycsw.core.metadata module
"""


from urllib.parse import parse_qs, urlparse

import pycsw.core.metadata
from owslib import util
from owslib.iso import MD_Metadata, namespaces
from pycsw.core.metadata import _set

# we need to access private protected members from pycsw.* package
# pylint: disable=protected-access

_parse_iso = pycsw.core.metadata._parse_iso


def rpi_parse_iso(context, repos, exml):

    """
    Patched version of pycsw.core.metadata._parse_iso function

    Added support for parsing new queryables
    """

    # create owslib object for ISO metadata
    metadata = MD_Metadata(exml)

    # parse by original
    recobj = _parse_iso(context, repos, exml)

    # add pycsw:RPIRecordUUID to parsed object
    # /gmd:MD_Metadata/@uuid is not parsed by owslib
    # MD_Metadata() we have to do it manually -> parsing xml
    _set(
        context,
        recobj,
        "pycsw:RPIRecordUUID",
        exml.attrib.get("uuid"),
    )

    # add pycsw:RPIOrganizationUUID to parsed object
    # gmd:contact/gmd:CI_ResponsibleParty/@uuid is not parsed by owslib
    # MD_Metadata() we have to do it manually -> parsing xml
    first_contact_el = exml.find(
        util.nspath_eval("gmd:contact[1]/gmd:CI_ResponsibleParty[1]", namespaces)
    )
    _set(
        context,
        recobj,
        "pycsw:RPIOrganizationUUID",
        first_contact_el.attrib.get("uuid"),
    )

    is_viewable = any(
        check_ows_url(online.url, "is_viewable")
        for online in metadata.distribution.online
    )

    is_searchable = any(
        check_ows_url(online.url, "is_searchable")
        for online in metadata.distribution.online
    )

    _set(context, recobj, "pycsw:RPIIsViewable", is_viewable)
    _set(context, recobj, "pycsw:RPIIsSearchable", is_searchable)

    parent_identifier_anchor_el = exml.find(
        util.nspath_eval("gmd:parentIdentifier/gmx:Anchor", namespaces)
    )

    if parent_identifier_anchor_el is not None:
        _set(
            context, recobj, "pycsw:ParentIdentifier", parent_identifier_anchor_el.text
        )

    dq_specification_title_anchor_el = exml.find(
        "gmd:dataQualityInfo/*/gmd:report/*/gmd:result/*/gmd:specification/*/gmd:title/gmx:Anchor",
        namespaces,
    )

    if dq_specification_title_anchor_el is not None:
        _set(
            context,
            recobj,
            "pycsw:SpecificationTitle",
            dq_specification_title_anchor_el.attrib.get(
                f"{{{namespaces['xlink']}}}href"
            ),
        )

    keyword_anchor_els = exml.findall(
        "gmd:identificationInfo/*/gmd:descriptiveKeywords/*/gmd:keyword/gmx:Anchor",
        namespaces,
    )

    anchored_keywords = []

    for keyword_anchor in keyword_anchor_els:
        if keyword_anchor is not None:
            anchored_keywords.append(
                keyword_anchor.attrib.get(f"{{{namespaces['xlink']}}}href")
            )

    if len(anchored_keywords) > 0:
        current_keywords = getattr(
            recobj, context.md_core_model["mappings"]["pycsw:Keywords"]
        )
        if current_keywords:
            current_keywords = current_keywords.split(",")
        else:
            current_keywords = []
        merged_keywords = ",".join(current_keywords + anchored_keywords)
        _set(context, recobj, "pycsw:Keywords", merged_keywords)

    return recobj


# monkey patching ISO metadata parsing function in pycsw package
pycsw.core.metadata._parse_iso = rpi_parse_iso


def check_ows_url(url: str, check_type: str):

    """
    Helper function checking if url is "viewable" or "searchable"
    OWS (WMS, WFS, WMTS) request
    """

    check_types = {
        "is_ows": False,
        "is_viewable": False,
        "is_searchable": False,
    }

    if check_type not in check_types:
        raise ValueError(f"Check type {check_type} not supported")

    # parse query params and normalize qparams names to lowercase
    qparams = {k.lower(): v for k, v in parse_qs(urlparse(url).query).items()}

    # request contains all the needed ows query params
    if {"service", "version", "request"}.issubset(qparams.keys()):
        check_types["is_ows"] = True

        if qparams["service"][0].lower() in ["wms", "wmts"]:
            if qparams["request"][0].lower() in ["getmap", "gettile"]:
                check_types["is_viewable"] = True

        if qparams["service"][0].lower() in ["wfs"]:
            if qparams["request"][0].lower() == "getfeature":
                check_types["is_searchable"] = True

    return check_types[check_type]



"""
Monkeypatches for pycsw.core.repository module
"""

import logging

import pycsw.core.repository
import requests
from sqlalchemy.sql import text

LOGGER = logging.getLogger(__name__)


def get_column_name(context, model_name):
    """
    Lookup function to get column name from mapping configuration (context)
    """
    return context.md_core_model["mappings"][model_name]


def execute_transaction_webhook(context, transaction_type, records):
    """
    Sends POST request o web hook URL (specified in config) with body
    containing summary about transaction executed
    """

    if (
        hasattr(context, "transactions_webhook_url")
        and context.transactions_webhook_url
    ):
        transactions = []

        for record in records:
            transaction = {
                "transaction_type": transaction_type,
                "record_id": getattr(
                    record, get_column_name(context, "pycsw:Identifier")
                ),
                "resource_type": getattr(
                    record, get_column_name(context, "pycsw:Type")
                ),
            }
            if transaction_type == "insert":
                transaction["record"] = getattr(
                    record, get_column_name(context, "pycsw:XML")
                )

            transactions.append(transaction)

        try:
            r = requests.post(
                context.transactions_webhook_url, json=transactions, timeout=10
            )
            r.raise_for_status()
        # pylint: disable=broad-exception-caught
        except Exception as e:
            LOGGER.warning("Notification to transactions web hook failed")
            LOGGER.warning(e)
    else:
        LOGGER.debug("Transactions web hook url not set, notification will not be send")


delete = pycsw.core.repository.Repository.delete
insert = pycsw.core.repository.Repository.insert


def rpi_insert(self, record, source, insert_date):
    """
    Patched version of pycsw.core.repository.insert function
    """

    # call original insert method
    insert(self, record, source, insert_date)

    # no exception has been raised, webhook may be called
    execute_transaction_webhook(self.context, "insert", [record])


def rpi_delete(self, constraint):
    """
    Patched version of pycsw.core.repository.delete function
    """

    # pylint: disable=protected-access
    records = (
        self._get_repo_filter(self.session.query(self.dataset))
        .filter(text(constraint["where"]))
        .params(self._create_values(constraint["values"]))
        .all()
    )

    # extend records with child records if any

    records.extend(
        self._get_repo_filter(self.session.query(self.dataset))
        .filter(
            getattr(
                self.dataset, get_column_name(self.context, "pycsw:ParentIdentifier")
            ).in_(
                [
                    getattr(record, get_column_name(self.context, "pycsw:Identifier"))
                    for record in records
                ]
            )
        )
        .all()
    )

    deleted = delete(self, constraint)

    # no exception has been raised, webhook may be called
    execute_transaction_webhook(self.context, "delete", records)

    return deleted


pycsw.core.repository.Repository.delete = rpi_delete
pycsw.core.repository.Repository.insert = rpi_insert
